import gzip
import numpy as np
import pyBigWig as pbw
import pandas as pd
class Feature():

    def __init__(self, **kwargs):
        self.load(**kwargs)
    
    def load(self):
        raise Exception('Not implemented')

    def get(self):
        raise Exception('Not implemented')

    def __len__(self):
        raise Exception('Not implemented')

class HiCFeature(Feature):

    def load(self, path = None):
        self.hic = self.load_hic(path)

    def get(self, start, window = 2097152, res = 10000):
        start_bin = int(start / res)
        range_bin = int(window / res)
        end_bin = start_bin + range_bin
        hic_mat = self.diag_to_mat(self.hic, start_bin, end_bin)
        return hic_mat

    def load_hic(self, path):
        # print(f'Reading Hi-C: {path}')
        return dict(np.load(path))

    def diag_to_mat(self, ori_load, start, end):
        '''
        Only accessing 256 x 256 region max, two loops are okay
        '''
        square_len = end - start
        diag_load = {}
        for diag_i in range(square_len):
            diag_load[str(diag_i)] = ori_load[str(diag_i)][start : start + square_len - diag_i]
            diag_load[str(-diag_i)] = ori_load[str(-diag_i)][start : start + square_len - diag_i]
        start -= start
        end -= start

        diag_region = []
        for diag_i in range(square_len):
            diag_line = []
            for line_i in range(-1 * diag_i, -1 * diag_i + square_len):
                if line_i < 0:
                    diag_line.append(diag_load[str(line_i)][start + line_i + diag_i])
                else:
                    diag_line.append(diag_load[str(line_i)][start + diag_i])
            diag_region.append(diag_line)
        diag_region = np.array(diag_region).reshape(square_len, square_len)
        return diag_region

    def __len__(self):
        return len(self.hic['0'])

class GenomicFeatureSingleThread(Feature):

    def __init__(self, path, norm):
        self.path = path
        self.load(path)
        self.norm = norm
        # print(f'Feature path: {path} \n Normalization status: {norm}')

    def load(self, path):
        self.feature = self.read_feature(path)

    def get(self, chr_name, start, end):
        feature = self.feature_to_npy(chr_name, start, end)
        feature = np.nan_to_num(feature, 0) # Important! replace nan with 0
        if self.norm == 'log':
            feature = np.log(feature + 1)
        elif self.norm is None:
            feature = feature
        else:
            raise Exception(f'Norm type {self.norm} undefined')
        return feature

    def read_feature(self, path):
        '''
        read bigwig file
        '''
        bw_file = pbw.open(path)
        return bw_file

    def feature_to_npy(self, chr_name, start, end):
        signals = self.feature.values(chr_name, start, end)
        return np.array(signals)

    def length(self, chr_name):
        return self.feature.chroms(chr_name)

class GenomicFeature(GenomicFeatureSingleThread):

    def __init__(self, path, norm):
        self.path = path
        self.norm = norm
        # print(f'Feature path: {path} \n Normalization status: {norm}')

    def load(self, path):
        raise Exception('Left blank')

    def feature_to_npy(self, chr_name, start, end):
        with pbw.open(self.path) as bw_file:
            signals = bw_file.values(chr_name, int(start), int(end))
        return np.array(signals)

    def length(self, chr_name):
        with pbw.open(self.path) as bw_file:
            length = bw_file.chroms(chr_name)
        return length

class SequenceFeature(Feature):

    def load(self, path = None):
        self.seq = self.read_seq(path)

    def get(self, start, end):
        seq = self.seq_to_npy(self.seq, start, end)
        onehot_seq = self.onehot_encode(seq)
        return onehot_seq

    def __len__(self):
        return len(self.seq)

    def read_seq(self, dna_dir):
        '''
        Transform fasta data to numpy array
        
        Args:
            dna_dir (str): Directory to DNA .fa path

        Returns:
            array: A numpy char array that contains DNA for a chromosome
        '''
        # print(f'Reading sequence: {dna_dir}')
        with gzip.open(dna_dir, 'r') as f:
            seq = f.read().decode("utf-8")
        seq = seq[seq.find('\n'):]
        seq = seq.replace('\n', '').lower()
        return seq
        
    def seq_to_npy(self, seq, start, end):
        '''
        Transform fasta data to integer numpy array
        
        Args:
            dna_dir (str): Directory to DNA .fa path

        Returns:
            array: A numpy char array that contains DNA for a chromosome
        '''
        seq = seq[start : end]
        en_dict = {'a' : 0, 't' : 1, 'c' : 2, 'g' : 3, 'n' : 4}
        en_seq = [en_dict[ch] for ch in seq]
        np_seq = np.array(en_seq, dtype = int)
        return np_seq

    def onehot_encode(self, seq):
        ''' 
        encode integer dna array to onehot (n x 5)
        Args:
            seq (arr): Numpy array (n x 1) of dna encoded as 0-4 integers

        Returns:
            array: A numpy matrix (n x 5)
        '''
        seq_emb = np.zeros((len(seq), 5))
        seq_emb[np.arange(len(seq)), seq] = 1
        return seq_emb

class InsulationScore(Feature):
    def load(self, path = None):
        self.insulation = self.read_table(path)
        print(f'Reading Insulation Table: {path}')
    
    def read_table(self, path):
        insulation_table = pd.read_csv(path)
        return insulation_table
    
    def get(self, chr_name, start_value, length=209, window=10, resolution=10000):
        end_value = start_value + length * resolution
        # print(chr_name, start_value, end_value)
        cache_insulation = self.insulation[self.insulation['chrom'] == chr_name]
        start_index = cache_insulation[cache_insulation['start'] == start_value].index[0]
        end_index = cache_insulation[cache_insulation['end'] == end_value].index[0]
        # print(start_index, end_index)
        # print(cache_insulation)
        extracted_rows = cache_insulation.loc[start_index:end_index]
        # print(extracted_rows)
        window_size = window * resolution
        insulation_score = np.array(extracted_rows[f'log2_insulation_score_{window_size}'])
        insulation_score = np.nan_to_num(insulation_score, 0)
        return insulation_score
    
    def __len__(self):
        return len(self.insulation)
    



#!/usr/bin/env python3
import argparse
import torch
from torch.utils.data import DataLoader
import os
import numpy as np
import time
import cooler
from hicompass.chromosome_dataset_predict import ChromosomeDataset
from hicompass.HiCompass_models import ConvTransModel
from hicompass.utils import chrome_size_dict, merge_hic_segment, pseudo_weight


def configure_parser(parser):
    """
    Configure the argument parser for the predict command
    """
    # Required arguments
    parser.add_argument('--cell_type', type=str, required=True,
                        help='Cell type name for the prediction')
    parser.add_argument('--atac_path', type=str, required=True,
                        help='Path to the ATAC-seq .bw file')
    parser.add_argument('--ctcf_path', type=str, required=True,
                        help='Path to the general CTCF ChIP-seq .bw file')
    parser.add_argument('--dna_dir_path', type=str, required=True,
                        help='Path to the DNA sequence fa.gz dir file')
    parser.add_argument('--omit_regions_path', type=str, required=True,
                        help='Path to the omit_regions csv file')
    parser.add_argument('--model_path', type=str, required=True,
                        help='Path to the model weights')

    # Optional arguments with defaults
    parser.add_argument('--output_path', type=str,
                        help='Output directory path (default: ./output/<cell_type>)')
    parser.add_argument('--device', type=str, default='cpu',
                        help='decide gpu or cpu usage (e.g., "cuda:0" or "cpu"), default cpu')
    parser.add_argument('--chromosomes', type=str, default='1-22',
                        help='Chromosomes to process (e.g., "1,3,5" or "1-22")')
    parser.add_argument('--depth', type=float, default=80e4,
                        help='Sequencing depth parameter')
    parser.add_argument('--stride', type=int, default=50,
                        help='Stride value for prediction (default: 50)')
    parser.add_argument('--batch_size', type=int, default=2,
                        help='Batch size for prediction (default: 2)')
    parser.add_argument('--num_workers', type=int, default=16,
                        help='Number of worker threads for data loading (default: 16)')
    parser.add_argument('--resolution', type=int, default=10000,
                        help='Resolution of the output Hi-C matrix in bp (default: 10000)')
    parser.add_argument('--window_size', type=int, default=2097152,
                        help='Window size for prediction in bp (default: 2097152)')

    return parser


def parse_chromosome_input(chrom_str):
    """Parse chromosome input to get list of chromosomes."""
    chromosomes = []

    if '-' in chrom_str:
        # Range format: "1-22"
        start, end = map(int, chrom_str.split('-'))
        chromosomes = [f'chr{i}' for i in range(start, end + 1)]
    else:
        # List format: "1,2,3,X,Y"
        chrom_list = chrom_str.split(',')
        chromosomes = [f'chr{c}' for c in chrom_list]

    return chromosomes


def run(args):
    """Run prediction with the specified arguments."""
    # Configure torch multiprocessing
    torch.multiprocessing.set_sharing_strategy('file_system')

    # Set device
    device = torch.device(args.device if torch.cuda.is_available() and "cuda" in args.device else "cpu")
    print(f"Using device: {device}")

    # Parse chromosomes
    chr_list = parse_chromosome_input(args.chromosomes)
    print(f"Processing chromosomes: {', '.join(chr_list)}")

    # Set output path
    if args.output_path is None:
        output_dir = f"./output/{args.cell_type}"
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        save_path = f"{output_dir}/{args.cell_type}.cool"
    else:
        output_dir = args.output_path
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        save_path = f"{output_dir}/{args.cell_type}.cool"

    print(f"Output will be saved to: {save_path}")

    # Load model
    print(f"Loading model from {args.model_path}")
    model = ConvTransModel()
    model.to(device)
    checkpoint = torch.load(args.model_path, map_location=device)
    model.load_state_dict(checkpoint, strict=True)
    model.eval()

    # Create dataset
    print(f"Creating dataset with ATAC-seq data from {args.atac_path}")
    start_time = time.time()
    dataset = ChromosomeDataset(
        chr_name_list=chr_list,
        atac_path_list=[args.atac_path],
        dna_dir_path=args.dna_dir_path,
        omit_regions_file_path=args.omit_regions_path,
        general_ctcf_bw_path=args.ctcf_path,
        stride=args.stride,
        depth=args.depth,
        use_aug=False
    )
    print(f"Dataset loaded in {time.time() - start_time:.2f} seconds")

    # Create dataloader
    dataloader = DataLoader(
        dataset,
        batch_size=args.batch_size,
        shuffle=False,
        num_workers=args.num_workers
    )

    # Initialize output dictionary
    output_dict = {chr_name: [] for chr_name in chr_list}

    # Run prediction
    start_time = time.time()
    with torch.no_grad():
        for step, data in enumerate(dataloader):
            if step % 10 == 0:
                print(f"Processing batch {step}/{len(dataloader)}")

            seq, atac, real_depth, ctcf, start, start_ratio, end, end_ratio, chr_name, chr_name_ratio = data

            # Forward pass
            mat_pred, pred_cls, insulation = model(
                seq.to(device),
                atac.to(device),
                real_depth.to(device),
                ctcf.to(device),
                start_ratio.to(device),
                end_ratio.to(device),
                chr_name_ratio.to(device)
            )

            # Process results
            for i in range(seq.shape[0]):
                result = mat_pred[i].cpu().detach().numpy()
                result = np.clip(result, a_max=10, a_min=0) * 10
                result = np.triu(result)
                np.fill_diagonal(result, 0)
                output_dict[str(chr_name[i])].append([start[i].cpu(), end[i].cpu(), result])

    print(f"Prediction completed in {time.time() - start_time:.2f} seconds")

    # Merge segments and save
    start_time = time.time()
    print(f"Merging segments and saving to {save_path}")
    merge_hic_segment(
        output_dict,
        save_path=save_path,
        window_size=args.window_size,
        resolution=args.resolution
    )
    print(f"Merging and saving completed in {time.time() - start_time:.2f} seconds")

    # Add pseudo weight
    pseudo_weight(save_path, save_path, weight=1)
    print(f'Pseudo weight=1 added')
    print(f"All operations completed. Results saved to {save_path}")


#!/usr/bin/env python3
import argparse
import sys
from hicompass.commands import predict

def main():
    """
    Hi-Compass main command
    """
    parser = argparse.ArgumentParser(
        description='Hi-Compass: A Hi-C predict deep learning model',
        usage='hicompass <command> [<args>]'
    )
    
    # 添加子命令解析器
    subparsers = parser.add_subparsers(dest='command', help='commands that you can use')
    
    # 添加predict子命令
    predict_parser = subparsers.add_parser('predict', help='predict a Hi-C cool result from input data')
    predict.configure_parser(predict_parser)
    
    # 解析命令行参数
    args = parser.parse_args()
    
    # 如果没有指定子命令，显示帮助信息
    if args.command is None:
        parser.print_help()
        sys.exit(1)
    
    # 根据子命令执行相应的功能
    if args.command == 'predict':
        predict.run(args)

if __name__ == "__main__":
    main()


import sys 
import os
import random
import pickle
import pandas as pd
import numpy as np

from skimage.transform import resize
from torch.utils.data import Dataset

import hicompass.data_feature as data_feature

cell_line_class = {'a549': 0, 'imr90': 1, 'h1_hesc': 2, 'adipocyte': 3, 
              'transverse_colon': 4, 'k562': 5, 'hffc6': 6, 'gm12878': 7, 
              'sun449': 8,'retina_1': 9,'retina_2':9, 'retina_3': 9, 'retina_4' : 9, 
              'gastrocnemius': 10, 'osteogenesis':11, 'neuron':12, 'huh7':13, 
              'huh1': 14, 'hmec': 15, 'hep3b': 16, 'dnd41': 17, 
              'ct27_stem' : 18, 'ct27_evt': 18, 
              'colon_5':19, 'colon_4':19, 'colon_3':19, 'colon_2':19, 'colon_1':19, 
              'adipogenesis':20, 'foresking_fibroblasts': 21, 'Tcell': 22}
chr_name_class = {'chr1':1, 'chr2': 2, 'chr3': 3, 'chr4': 4,
                  'chr5': 5, 'chr6': 6, 'chr7': 7, 'chr8': 8, 'chr9': 9,
                  'chr10': 10, 'chr11': 11, 'chr12': 12, 'chr13': 13, 'chr14': 14,
                  'chr15': 15, 'chr16': 16, 'chr17': 17, 'chr18': 18, 'chr19': 19,
                  'chr20': 20, 'chr21': 21, 'chr22': 22, 'chrX': 23, 'chrY': 24}
chrome_size_dict = {'chr1': 248956422, 'chr2': 242193529, 'chr3': 198295559, 'chr4': 190214555,
 'chr5': 181538259, 'chr6': 170805979, 'chr7': 159345973, 'chr8': 145138636, 
 'chr9': 138394717, 'chr10': 133797422, 'chr11': 135086622, 'chr12': 133275309, 
 'chr13': 114364328, 'chr14': 107043718, 'chr15': 101991189, 'chr16': 90338345, 
 'chr17': 83257441, 'chr18': 80373285, 'chr19': 58617616, 'chr20': 64444167, 'chr21': 46709983, 
 'chr22': 50818468, 'chrX': 156040895, 'chrY': 57227415}

class ChromosomeDataset(Dataset):
    '''
    Dataloader that provide sequence, features, and HiC data. Assume input
    folder strcuture.
    Args:
        data_root (str): Directory including sequence features, and HiC matrix 
            as subdirectories.
        chr_name (str): Name of the represented chromosome (e.g. chr1)
            as ``root/DNA/chr1/DNA`` for DNA as an example.
        omit_regions (list of tuples): start and end of excluded regions
    '''
    def __init__(self, chr_name_list, atac_path_list, depth, 
        general_ctcf_bw_path,
        dna_dir_path, 
        omit_regions_file_path,
        chrom_max_num=24,
        use_aug=False, stride=158):
        self.use_aug = use_aug
        self.res = 10000 # 10kb resolution
        self.bins = 209.7152 # 209.7152 bins 2097152 bp
        self.sample_bins = 500
        self.chrom_max_num = chrom_max_num
        self.atac_path_list = atac_path_list
        self.dna_dir_path = dna_dir_path
        self.general_ctcf_bw_path = general_ctcf_bw_path
        self.stride = stride 
        self.depth = depth
        self.chr_name_list = chr_name_list
        self.omit_regions_dict = proc_centrotelo(omit_regions_file_path)
        print(f'Loading chromosome {self.chr_name_list}...')
        self.seq_dict = {}
        for chr_name in self.chr_name_list:
            self.seq_dict[chr_name] = data_feature.SequenceFeature(path=f'{self.dna_dir_path}/{chr_name}.fa.gz')
        
        self.atac_list = [data_feature.GenomicFeature(path=path, norm=None) for path in self.atac_path_list]
        self.ctcf_feature = data_feature.GenomicFeature(path=self.general_ctcf_bw_path, norm= None)
        self.all_intervals_list = []
        for chr_name in self.chr_name_list:
            self.all_intervals_list.append(self.get_intervals_chr_ct_dep(seq=self.seq_dict[chr_name],
                                                                            chr_name=chr_name,
                                                                            omit_regions=self.omit_regions_dict[chr_name]))

        self.intervals = np.concatenate(self.all_intervals_list, axis=0)


    def __getitem__(self, idx):
        start, end, chr_name = self.intervals[idx]
        # start, end, chr_name = self.intervals[idx]
        start = int(start)
        end = int(end)
        depth = int(self.depth)
        target_size = int(self.bins * self.res)
        total_seq = self.seq_dict[chr_name]
        total_atac_list = self.atac_list
        total_ctcf = self.ctcf_feature
        start, end = self.shift_fix(target_size, start)
        start = int(start)
        end = int(end)
        seq, atac, real_depth, ctcf = self.get_data_at_chr_interval(start=start, end=end, chr_name=chr_name,
                                                                total_seq=total_seq,
                                                                atac_list=total_atac_list,
                                                                ctcf=total_ctcf,
                                                                depth=depth)
        return seq, atac, real_depth, ctcf, start, start/chrome_size_dict[chr_name], end, end/chrome_size_dict[chr_name], chr_name, chr_name_class[chr_name]/self.chrom_max_num

    def __len__(self):
        return len(self.intervals)


    def get_data_at_chr_interval(self, start, end, chr_name, total_seq, atac_list, ctcf, depth):
        '''
        used in get item
        '''
        # Sequence processing
        start = int(start)
        end = int(end)
        # print('get_data_at_chr_interval',start, end)
        seq = total_seq.get(start, end)
        # Features processing
        features_atac_list = [item.get(chr_name, start, end) for item in atac_list]
        features_atac = np.sum(np.array(features_atac_list), axis=0)
        # features_atac = atac_list[0].get(chr_name, start, end)
        features_ctcf = ctcf.get(chr_name, start, end)
        real_depth = int(depth)
        real_depth = min(len(features_ctcf), real_depth)
        return seq, features_atac, real_depth, features_ctcf

    
    def get_intervals_chr_ct_dep(self, seq, chr_name, omit_regions):
        # chr_num = int(chr_name.split('chr')[-1])
        chr_bins = len(seq) / self.res
        data_size = (chr_bins - self.sample_bins) / self.stride
        starts = np.arange(0, data_size).reshape(-1, 1) * self.stride
        intervals_bin = np.append(starts, starts + self.sample_bins, axis=1)
        intervals = intervals_bin * self.res
        intervals = intervals.astype(int)
        intervals = self.filter(intervals, omit_regions)
        chr_name_array = np.full(len(intervals), chr_name).reshape(-1, 1)
        intervals = np.append(intervals, chr_name_array, axis=1)
        return intervals
    

    def get_active_intervals(self):
        '''
        Get intervals for sample data: [[start, end]]
        '''
        chr_bins = len(self.seq) / self.res
        data_size = (chr_bins - self.sample_bins) / self.stride
        starts = np.arange(0, data_size).reshape(-1, 1) * self.stride
        intervals_bin = np.append(starts, starts + self.sample_bins, axis=1)
        intervals = intervals_bin * self.res
        return intervals.astype(int)

    def filter(self, intervals, omit_regions):
        valid_intervals = []
        for start, end in intervals: 
            start_cond = start <= omit_regions[:, 1]
            end_cond = omit_regions[:, 0] <= end
            if sum(start_cond * end_cond) == 0:
                valid_intervals.append([start, end])
        return valid_intervals

    def encode_seq(self, seq):
        ''' 
        encode dna to onehot (n x 5)
        '''
        seq_emb = np.zeros((len(seq), 5))
        seq_emb[np.arange(len(seq)), seq] = 1
        return seq_emb


    def shift_fix(self, target_size, start):
        offset = 0
        return start + offset , start + offset + target_size
    


def proc_centrotelo(bed_dir):
    ''' Take a bed file indicating location, output a dictionary of items 
    by chromosome which contains a list of 2 value lists (range of loc)
    '''
    df = pd.read_csv(bed_dir , sep='\t', names=['chr', 'start', 'end'])
    chrs = df['chr'].unique()
    centrotelo_dict = {}
    for chr_name in chrs:
        sub_df = df[df['chr'] == chr_name]
        regions = sub_df.drop('chr', axis = 1).to_numpy()
        centrotelo_dict[chr_name] = regions
    return centrotelo_dict







import numpy as np
import os
import pandas as pd
from functools import reduce
import time
import cooler
chrome_size_dict = {'chr1': 248956422, 'chr2': 242193529, 'chr3': 198295559, 'chr4': 190214555,
 'chr5': 181538259, 'chr6': 170805979, 'chr7': 159345973, 'chr8': 145138636, 
 'chr9': 138394717, 'chr10': 133797422, 'chr11': 135086622, 'chr12': 133275309, 
 'chr13': 114364328, 'chr14': 107043718, 'chr15': 101991189, 'chr16': 90338345, 
 'chr17': 83257441, 'chr18': 80373285, 'chr19': 58617616, 'chr20': 64444167, 'chr21': 46709983, 
 'chr22': 50818468}


def positive(matrix):
    min_value = matrix.min()
    if min_value>=0:
        return matrix
    else:
        return matrix-min_value

def norm_hic(input_matrix, max_num=6):
    now_max = max(np.diagonal(input_matrix))
    return input_matrix/((now_max+0.0001)/(max_num+0.0001))


def make_cooler_bins_chr(chr_name, length, resolution):
    total_bin = []
    for i in range(0, length, resolution):
        total_bin.append([chr_name, i, min(length, i+resolution)])
    df = pd.DataFrame(total_bin, columns=['chrom', 'start', 'end'])
    return df

def get_chr_stack(chr_list, chr_name, chrome_size_dict=chrome_size_dict, resolution=10000):
    chrome_size_dict = {key: chrome_size_dict[key] for key in chr_list if key in chrome_size_dict}
    chr_before_list = chr_list[:chr_list.index(chr_name)]
    if len(chr_before_list) == 0:
        return 0
    else:
        stack = 0
        for before_chr in chr_before_list:
            stack+= int(chrome_size_dict[before_chr]/resolution) + 1
        return stack

def merge_hic_segment(hic_list, save_path, window_size=2097152, resolution=10000):
    chr_list = [i for i in hic_list]
    bins = int(window_size/resolution)
    chr_hic_dict = {}
    for chr_num in chr_list:
        chr_hic_dict[chr_num] = [make_cooler_bins_chr(chr_name=chr_num, length=chrome_size_dict[chr_num], resolution=resolution)]
        sub_list = hic_list[chr_num]
        # print(sub_list)
        large_pic = np.zeros((int(chrome_size_dict[chr_num]/resolution), 
                              int(chrome_size_dict[chr_num]/resolution)))
        for segement in sub_list:
            # print(segement[0])
            sub_start_bin = int(segement[0]/resolution)
            large_pic[sub_start_bin: sub_start_bin+bins, sub_start_bin: sub_start_bin+bins] = segement[2]
        # 将单chr的矩阵稀疏化
        rows, cols = np.nonzero(large_pic)
        stack = get_chr_stack(chr_list=chr_list, chr_name=chr_num, chrome_size_dict=chrome_size_dict, resolution=resolution)
        rows += stack
        cols += stack
        counts = large_pic[np.nonzero(large_pic)]
        large_pic_sp = np.column_stack((rows, cols, counts))
        large_pic_sp = pd.DataFrame(large_pic_sp)
        chr_hic_dict[chr_num].append(large_pic_sp)
    # 每个key对应的是一个list, [bin表, 单chr的hic三列矩阵]
    total_bin_list = []
    total_sp_list = []
    start_bin = 0
    total_bin_len = 0
    for chr_num in chr_hic_dict:
        total_bin_list.append(chr_hic_dict[chr_num][0])
        total_sp_list.append(chr_hic_dict[chr_num][1])
        # total_bin_len += chr_hic_dict[chr_num][1].shape[0]
    bin_df = reduce(lambda x, y: pd.concat([x,y], axis = 0), total_bin_list)
    bin_df.reset_index()
    bin_df.columns = ['chrom', 'start', 'end']
    sp_df = reduce(lambda x, y: pd.concat([x,y], axis = 0), total_sp_list)
    sp_df.reset_index()
    sp_df.columns = ['bin1_id', 'bin2_id', 'count']
    cooler.create_cooler(cool_uri=save_path, 
                         bins=bin_df, pixels=sp_df)

def cal_depth(filename):
    with open(filename, 'r') as file:
        depth = 0
        for _ in file:
            depth += 1
    return depth

def cal_size(filename):
    stats = os.stat(filename)
    return stats.st_size  


def pseudo_weight(mcool_path, save_path, weight=1):
    """
    mcool_path: path to mcool for processing
    res: resolution
    save_path: path to saved cool file
    weight: the pseudo weight
    """
    c = cooler.Cooler(f'{mcool_path}')
    bins_df = c.bins()[:]
    bins_df['weight'] = weight
    cooler.create_cooler(cool_uri=f'{save_path}', bins=bins_df, pixels=c.pixels()[:])
    cooler_obj = cooler.Cooler(f'{save_path}')
    stats = {
    "min_nnz": 0,
    "min_count": 0,
    "mad_max": 0,
    "cis_only": False,
    "ignore_diags": 2,
    "converged": False,
    "divisive_weights": False,
    }
    with cooler_obj.open("r+") as cljr:
        cljr["bins"]['weight'].attrs.update(stats)


import torch
from torch.utils.data import DataLoader
from hicompass.chromosome_dataset_predict import ChromosomeDataset
import os
import pandas as pd
from hicompass.HiCompass_models import ConvTransModel
import numpy as np
from skimage.transform import resize
from functools import reduce
import time
# time.sleep(1800 * 8)
from hicompass.utils import chrome_size_dict, make_cooler_bins_chr, get_chr_stack, merge_hic_segment, cal_depth, cal_size
torch.multiprocessing.set_sharing_strategy('file_system')
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu") # 推理就只用单卡就行

chr_list = ['chr'+ str(i) for i in range(1, 23)]
stride = 50 # keep 50 diag
cell_type= 'Tcell'
cell_type = 'CD8_Naive~2'
cell_type_name = cell_type.split('~')[0]
cell_type_list = [
    # 'transverse_colon', 'sun449', 'osteogenesis', 'k562', 'imr90', 'huh7','huh1', 'hmec', 'hffc6', 'hep3b', 
                  'h1_hesc','gm12878', 'gastrocnemius', 'foresking_fibroblasts', 'dnd41', 'ct27_stem', 'ct27_evt', 'a549'
                  ]
for cell_type_name in cell_type_list:
    if not os.path.isdir(f'/cluster/home/Yuanchen/project/scHiC/dataset/ATAC/{cell_type_name}'):
        continue
    model_weight_path = '/cluster/home/Yuanchen/project/scHiC/weight/TJ/cross/model_2.pth'
    if not os.path.exists(f'/cluster/home/Yuanchen/project/scHiC/paper_figure/fig2/cell_line_predict_result_80/{cell_type_name}'):
        os.system(f'mkdir /cluster/home/Yuanchen/project/scHiC/paper_figure/fig2/cell_line_predict_result_80/{cell_type_name}/')
    save_path = f'/cluster/home/Yuanchen/project/scHiC/paper_figure/fig2/cell_line_predict_result_80/{cell_type_name}/{cell_type_name}.cool'
    bw_path = f"/cluster/home/Yuanchen/project/scHiC/dataset/ATAC/{cell_type_name}/{cell_type_name}_ATAC_80e4/{cell_type_name}_ATAC_80e4.bw"
    # bw_path = f'/cluster/home/Yuanchen/project/scHiC/dataset/scATAC/output/atac_v1_pbmc_10k/outs/hic_cluster_n=10_new/{cell_type}/{cell_type}.bw'
    depth = 80e4


    model = ConvTransModel()
    model.to(device)
    checkpoint = torch.load(model_weight_path, map_location=device)
    model_weights = checkpoint
    model.load_state_dict(model_weights, strict=True)
    model.eval()

    cache_barcode_path_list = [bw_path]
    startTime = time.time()
    dataset = ChromosomeDataset(chr_name_list=chr_list, 
                                    atac_path_list=cache_barcode_path_list,
                                    stride=stride, 
                                    depth=depth,use_aug=False)
    print("load dataset in %f s" % (time.time() - startTime))
    dataloader = DataLoader(dataset, batch_size=2, shuffle=False, num_workers=16)
    output_dict = {}
    for chr_name in chr_list:
        output_dict[chr_name] = []
    startTime = time.time()
    for step, data in enumerate(dataloader):
        seq, atac, real_depth, ctcf, start, start_ratio, end, end_ratio, chr_name, chr_name_ratio = data
        sub_time = time.time()
        mat_pred, pred_cls, insulation = model(seq.to(device), atac.to(device), real_depth.to(device), ctcf.to(device), start_ratio.to(device), end_ratio.to(device), chr_name_ratio.to(device)) 
        for i in range(seq.shape[0]):
            result = mat_pred[i].cpu().detach().numpy()
            result = np.clip(result, a_max=10, a_min=0) * 10
            result = np.triu(result)
            np.fill_diagonal(result, 0)
            output_dict[str(chr_name[i])].append([start[i].cpu(), end[i].cpu(), result])
    print("Run model in %f s" % (time.time() - startTime))
    startTime = time.time()
    merge_hic_segment(output_dict,save_path=save_path, window_size=2097152, resolution=10000)
    print("Run merge in %f s" % (time.time() - startTime))


import torch
import torch.nn as nn

import hicompass.blocks as blocks


class ConvModel(nn.Module):
    def __init__(self, num_genomic_features, mid_hidden=256):
        super(ConvModel, self).__init__()
        print('Initializing ConvModel')
        self.encoder = blocks.EncoderSplit(num_genomic_features, output_size=mid_hidden, num_blocks=12)
        self.decoder = blocks.Decoder(mid_hidden * 2)
        self.adpool = nn.AdaptiveAvgPool2d((209, 209))

    def forward(self, x):
        '''
        Input feature:
        batch_size, length * res, feature_dim
        '''
        x = self.move_feature_forward(x).float()
        x = self.encoder(x)
        x = self.diagonalize(x)
        x = self.decoder(x).squeeze(1)
        x = self.adpool(x)
        return x

    def move_feature_forward(self, x):
        '''
        input dim:
        bs, img_len, feat
        to: 
        bs, feat, img_len
        '''
        return x.transpose(1, 2).contiguous()

    def diagonalize(self, x):
        # print("x::", x.shape)
        x_i = x.unsqueeze(2).repeat(1, 1, 256, 1)
        # print("x_i::", x_i.shape)
        x_j = x.unsqueeze(3).repeat(1, 1, 1, 256)
        # print("x_j::", x_j.shape)
        input_map = torch.cat([x_i, x_j], dim=1)
        # print("input_map::", input_map.shape)
        return input_map


class ConvTransModel(nn.Module):
    def __init__(self, encoder_split=blocks.EncoderSplit(), num_classes=30, mid_hidden=256, record_attn=False):
        super(ConvTransModel, self).__init__()
        print('Initializing Discriminator ConvTransModel')
        self.encoder_split = encoder_split
        self.attn = blocks.AttnModule(hidden=mid_hidden, record_attn=record_attn)
        self.decoder = blocks.Decoder(mid_hidden*2)
        self.record_attn = record_attn
        self.adpool = nn.AdaptiveAvgPool2d((209, 209))
        
        # Discriminator branch
        self.pos_embed = nn.Linear(3, 12)
        self.resnet = blocks.ResNet(blocks.ResBlock, [2, 2, 2, 2], num_classes=num_classes)
        self.fc = nn.Linear(12 + num_classes, num_classes)

        # InsulationScoreModule
        self.ismodule = blocks.InsulationScoreModule()

    def forward(self, seq, atac, real_depth, ctcf, pos_start, pos_end, chrom_num):
        # 通过编码器获取融合后的特征表示
        seq = self.move_feature_forward(seq).float()
        fused_feat = self.encoder_split(seq=seq, atac=atac, real_depth=real_depth, ctcf=ctcf)
        # print(fused_feat.shape)
        fused_feat = self.move_feature_forward(fused_feat)
        if self.record_attn:
            fused_feat, attn_weights = self.attn(fused_feat)
        else:
            fused_feat = self.attn(fused_feat)
        fused_feat = self.move_feature_forward(fused_feat)
        fused_feat = self.diagonalize(fused_feat)
        # print(fused_feat.shape)
        x = self.decoder(fused_feat).squeeze(1)
        x = self.adpool(x)
        
        # Ignore diagonal
        diag_mask = torch.ones_like(x)
        diag_mask = torch.tril(diag_mask, 0) * torch.triu(diag_mask, -0)
        diag_mask = torch.where(diag_mask == 0, 1, 0) 
        x = diag_mask * x
        x = torch.clamp(x,0,10,out=None) + torch.eye(x.shape[1], dtype=x.dtype, device=x.device) * (1e-6)

        # InsulationScoreModule
        insulation = self.ismodule(x)

        # Discriminator
        mat = x.unsqueeze(1)
        start_pos = pos_start.unsqueeze(-1)
        end_pos = pos_end.unsqueeze(-1)
        chrom_num = chrom_num.unsqueeze(-1)
        pos_info = torch.cat([start_pos, end_pos, chrom_num], dim=1)
        pos_info = pos_info.float()
        # print(pos_info.shape)
        pos_feat = self.pos_embed(pos_info)
        mat_feat = self.resnet(mat)
        com_feat = torch.cat([pos_feat, mat_feat], dim=1)
        pred_cls = self.fc(com_feat)
        if self.record_attn:
            return x, pred_cls, insulation, attn_weights
        else:
            return x, pred_cls, insulation

    def move_feature_forward(self, x):
        return x.transpose(1, 2).contiguous()

    def diagonalize(self, x):
        x_i = x.unsqueeze(2).repeat(1, 1, 256, 1)
        x_j = x.unsqueeze(3).repeat(1, 1, 1, 256)
        input_map = torch.cat([x_i, x_j], dim=1)
        return input_map


# setup.py
import io
import os
from setuptools import setup, find_packages
# these things are needed for the README.md show on pypi (if you dont need delete it)
here = os.path.abspath(os.path.dirname(__file__))
# Import the README and use it as the long-description.
DESCRIPTION = 'Hi-Compass: A human hg38 cell-type specific Hi-C(10k for now) predict deep learning model'
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION
# Package meta-data.
NAME = 'hicompass'

EMAIL = '2247143021@qq.com'
URL="https://github.com/EndeavourSyc/Hi-Compass/"
AUTHOR = 'Yuanchen Sun'
VERSION = '0.1.7'
setup(
name=NAME,
version=VERSION,
author=AUTHOR,
author_email=EMAIL,
license='MIT',
description=DESCRIPTION,
url=URL,
long_description_content_type="text/markdown",
long_description=long_description,
packages=find_packages(),
install_requires=["numpy",
"pandas", "scikit-image", "pyBigWig", "cooler", ],   # PyTorch is not included as it needs to be installed manually according to the system configuration
classifiers=[
'Programming Language :: Python',
"Programming Language :: Python :: 3",
"License :: OSI Approved :: MIT License",
"Operating System :: OS Independent",],
python_requires=">=3.8",
    entry_points={"console_scripts": ["hicompass=hicompass.cli:main",],}
)


import torch
import torch.nn as nn
import numpy as np
import copy


class ConvBlock(nn.Module):
    def __init__(self, size, stride=2, hidden_in=64, hidden=64):
        super(ConvBlock, self).__init__()
        pad_len = int(size / 2)
        self.scale = nn.Sequential(
            nn.Conv1d(hidden_in, hidden, size, stride, pad_len),
            nn.BatchNorm1d(hidden),
            nn.ReLU(),
        )
        self.res = nn.Sequential(
            nn.Conv1d(hidden, hidden, size, padding=pad_len),
            nn.BatchNorm1d(hidden),
            nn.ReLU(),
            nn.Conv1d(hidden, hidden, size, padding=pad_len),
            nn.BatchNorm1d(hidden),
        )
        self.relu = nn.ReLU()

    def forward(self, x):
        scaled = self.scale(x)
        identity = scaled
        res_out = self.res(scaled)
        out = self.relu(res_out + identity)
        return out
    


class ATACDepthEncoder(nn.Module):
    def __init__(self, in_channel=1, output_size=128, filter_size=5, num_blocks=12, depth_ranges=None):
        super(ATACDepthEncoder, self).__init__()
        self.filter_size = filter_size
        self.conv_start = nn.Sequential(
            nn.Conv1d(in_channel, 32, 3, 2, 1, bias=False).float(),
            nn.BatchNorm1d(32),
            nn.ReLU(),
        )
        self.conv_start_1d = nn.Conv1d(in_channel, 32, 3, 2, 1)
        self.bn1d = nn.BatchNorm1d(32)
        self.activate = nn.ReLU()
        sizes_3 = [3] * num_blocks
        hiddens = [32, 32, 32, 32, 64, 64, 128, 128, 128, 128, 256, 256]
        hidden_ins = [32, 32, 32, 32, 32, 64, 64, 128, 128, 128, 128, 256]
        self.res_blocks = self.get_res_blocks(num_blocks, sizes_3, hidden_ins, hiddens)

        self.conv_end = nn.Conv1d(256, output_size, 1)
        
        # 添加深度嵌入层
        if depth_ranges is None:
            depth_ranges = [200000, 500000, 800000, 1200000, 1700000, 2200000]
        self.depth_ranges = depth_ranges
        self.depth_embed = nn.Embedding(len(depth_ranges) + 1, 1)

    def forward(self, atac, depth):
        # 将深度值映射到区间索引
        depth_idx = torch.bucketize(depth, torch.tensor(self.depth_ranges, device=depth.device))
        
        # 将区间索引输入到嵌入层
        depth_feat = self.depth_embed(depth_idx)
        # depth_feat = depth_feat.unsqueeze(2).expand(-1, -1, atac.size(1))
        
        # 将ATAC序列与深度特征连接
        # print('atac shape: ', atac.shape)
        # print('depth_feat shape: ', depth_feat.shape)

        x =(atac * depth_feat).unsqueeze(1).to(torch.float32) # (atac * depth_feat).unsqueeze(1).double()
        # print(type(x),x.dtype)
        # print('start', x.shape)
        x = self.conv_start_1d(x) 
        # print('conv_start_1d', x.shape)
        x = self.bn1d(x)
        # print('bn1d', x.shape)
        x = self.activate(x)
        x = self.res_blocks(x)
        # print('res_blocks', x.shape)
        out = self.conv_end(x)
        # print('conv_end', x.shape)
        return out

    def get_res_blocks(self, n, sizes, his, hs):
        blocks = []
        for i, filter_size, h, hi in zip(range(n), sizes, hs, his):
            blocks.append(ConvBlock(filter_size, hidden_in=hi, hidden=h))
        res_blocks = nn.Sequential(*blocks)
        return res_blocks


class DNAEncoder(nn.Module):
    def __init__(self, in_channel=5, output_size=64, filter_size=5, num_blocks=12):
        super(DNAEncoder, self).__init__()
        self.filter_size = filter_size
        self.conv_start = nn.Sequential(
            nn.Conv1d(in_channel, 32, 3, 2, 1),
            nn.BatchNorm1d(32),
            nn.ReLU(),
        )
        sizes_3 = [3] * num_blocks
        
        hiddens = [32, 32, 32, 32, 64, 64, 128, 128, 128, 128, 256, 256]
        hidden_ins = [32, 32, 32, 32, 32, 64, 64, 128, 128, 128, 128, 256]
        self.res_blocks = self.get_res_blocks(num_blocks, sizes_3, hidden_ins, hiddens)
        self.conv_end = nn.Conv1d(256, output_size, 1)
        
        # 添加特征权重
        self.feature_weights = nn.Parameter(torch.ones(1, in_channel, 1))

    def forward(self, x):
        # 应用特征权重
        # print(x.shape)
        x = x * self.feature_weights
        
        x = self.conv_start(x)
        
        x = self.res_blocks(x)
        out = self.conv_end(x)
        return out
    def get_res_blocks(self, n, sizes, his, hs):
        blocks = []
        for i, filter_size, h, hi in zip(range(n), sizes, hs, his):
            blocks.append(ConvBlock(filter_size, hidden_in=hi, hidden=h))
        res_blocks = nn.Sequential(*blocks)
        return res_blocks



class CTCFEncoder(nn.Module):
    def __init__(self, in_channel=1, output_size=64, filter_size=5, num_blocks=12):
        super(CTCFEncoder, self).__init__()
        self.filter_size = filter_size
        self.conv_start = nn.Sequential(
            nn.Conv1d(in_channel, 32, 3, 2, 1).float(),
            nn.BatchNorm1d(32),
            nn.ReLU(),
        )
        self.conv_start_1d = nn.Conv1d(in_channel, 32, 3, 2, 1)
        self.bn1d = nn.BatchNorm1d(32)
        self.activate = nn.ReLU()
        sizes_3 = [3] * num_blocks

        hiddens = [32, 32, 32, 32, 64, 64, 128, 128, 128, 128, 256, 256]
        hidden_ins = [32, 32, 32, 32, 32, 64, 64, 128, 128, 128, 128, 256]
        self.res_blocks = self.get_res_blocks(num_blocks, sizes_3, hidden_ins, hiddens)

        self.conv_end = nn.Conv1d(256, output_size, 1)

    def forward(self, x):
        x = x.unsqueeze(1).to(torch.float32)
        x = self.conv_start_1d(x) 
        x = self.bn1d(x)
        x = self.activate(x)
        x = self.res_blocks(x)
        out = self.conv_end(x)
        return out

    def get_res_blocks(self, n, sizes, his, hs):
        blocks = []
        for i, filter_size, h, hi in zip(range(n), sizes, hs, his):
            blocks.append(ConvBlock(filter_size, hidden_in=hi, hidden=h))
        res_blocks = nn.Sequential(*blocks)
        return res_blocks

# seq, atac, real_depth, ctcf, mat
class EncoderSplit(nn.Module):
    def __init__(self, atac_depth_encoder=ATACDepthEncoder(), dna_encoder=DNAEncoder(), ctcf_encoder=CTCFEncoder(), fusion_method='attention'):
        super(EncoderSplit, self).__init__()
        self.atac_depth_encoder = atac_depth_encoder
        self.dna_encoder = dna_encoder
        self.ctcf_encoder = ctcf_encoder
        self.fusion_method = fusion_method

        if fusion_method == 'attention':
            self.attention = AttnModule(hidden=256, )
        elif fusion_method == 'gating':
            self.gate = nn.Sequential(
                nn.Linear(256 * 3, 256),
                nn.Sigmoid()
            )

    def forward(self, seq, atac, real_depth, ctcf):
        atac_depth_out = self.atac_depth_encoder(atac, real_depth)
        dna_out = self.dna_encoder(seq)
        ctcf_out = self.ctcf_encoder(ctcf)

        if self.fusion_method == 'attention':
            # print(atac_depth_out.shape, dna_out.shape, ctcf_out.shape)
            concat_out = torch.cat([atac_depth_out, dna_out, ctcf_out], dim=1)
            # print(concat_out.shape)
            concat_out = concat_out.permute(0, 2, 1)
            fused_out = self.attention(concat_out)
            fused_out = fused_out.permute(0, 2, 1)
            # print(fused_out.shape)
            # fused_out = fused_out.mean(dim=2)
        elif self.fusion_method == 'gating':
            concat_out = torch.cat([atac_depth_out, dna_out, ctcf_out], dim=1)
            gate_weights = self.gate(concat_out)
            fused_out = gate_weights * concat_out
        else:
            fused_out = torch.cat([atac_depth_out, dna_out, ctcf_out], dim=1)

        return fused_out


class ResBlockDilated(nn.Module):
    def __init__(self, size, hidden=64, stride=1, dil=2):
        super(ResBlockDilated, self).__init__()
        pad_len = dil
        self.res = nn.Sequential(
            nn.Conv2d(hidden, hidden, size, padding=pad_len,
                      dilation=dil),
            nn.BatchNorm2d(hidden),
            nn.ReLU(),
            nn.Conv2d(hidden, hidden, size, padding=pad_len,
                      dilation=dil),
            nn.BatchNorm2d(hidden),
        )
        self.relu = nn.ReLU()

    def forward(self, x):
        identity = x
        res_out = self.res(x)
        out = self.relu(res_out + identity)
        return out


class Decoder(nn.Module):
    def __init__(self, in_channel, hidden=256, filter_size=3, num_blocks=5):
        super(Decoder, self).__init__()
        self.filter_size = filter_size

        self.conv_start = nn.Sequential(
            nn.Conv2d(in_channel, hidden, 3, 1, 1),
            nn.BatchNorm2d(hidden),
            nn.ReLU(),
        )
        self.res_blocks = self.get_res_blocks(num_blocks, hidden)
        self.conv_end = nn.Conv2d(hidden, 1, 1)

    def forward(self, x):
        x = self.conv_start(x)
        x = self.res_blocks(x)
        out = self.conv_end(x)
        return out

    def get_res_blocks(self, n, hidden):
        blocks = []
        for i in range(n):
            dilation = 2 ** (i + 1)
            blocks.append(ResBlockDilated(self.filter_size, hidden=hidden, dil=dilation))
        res_blocks = nn.Sequential(*blocks)
        return res_blocks


class TransformerLayer(torch.nn.TransformerEncoderLayer):
    # Pre-LN structure

    def forward(self, src, src_mask=None, src_key_padding_mask=None):
        # MHA section
        src_norm = self.norm1(src)
        src_side, attn_weights = self.self_attn(src_norm, src_norm, src_norm,
                                                attn_mask=src_mask,
                                                key_padding_mask=src_key_padding_mask)
        src = src + self.dropout1(src_side)

        # MLP section
        src_norm = self.norm2(src)
        src_side = self.linear2(self.dropout(self.activation(self.linear1(src_norm))))
        src = src + self.dropout2(src_side)
        return src, attn_weights


class TransformerEncoder(torch.nn.TransformerEncoder):

    def __init__(self, encoder_layer, num_layers, norm=None, record_attn=False):
        super(TransformerEncoder, self).__init__(encoder_layer, num_layers)
        self.layers = self._get_clones(encoder_layer, num_layers)
        self.num_layers = num_layers
        self.norm = norm
        self.record_attn = record_attn

    def forward(self, src, mask=None, src_key_padding_mask=None):
        r"""Pass the input through the encoder layers in turn.

        Args:
            src: the sequence to the encoder (required).
            mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """
        output = src

        attn_weight_list = []

        for mod in self.layers:
            output, attn_weights = mod(output, src_mask=mask, src_key_padding_mask=src_key_padding_mask)
            attn_weight_list.append(attn_weights.unsqueeze(0).detach())
        if self.norm is not None:
            output = self.norm(output)

        if self.record_attn:
            return output, torch.cat(attn_weight_list)
        else:
            return output

    def _get_clones(self, module, N):
        return torch.nn.modules.ModuleList([copy.deepcopy(module) for i in range(N)])


class PositionalEncoding(nn.Module):

    def __init__(self, hidden, dropout=0.1, max_len=256):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, hidden, 2) * (-np.log(10000.0) / hidden))
        pe = torch.zeros(max_len, 1, hidden)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    def forward(self, x):
        """
        Args:
            x: Tensor, shape [seq_len, batch_size, embedding_dim]
        """
        x = x + self.pe[:x.size(0)]
        return self.dropout(x)


class AttnModule(nn.Module):
    def __init__(self, hidden=256, layers=8, record_attn=False, inpu_dim=256):
        super(AttnModule, self).__init__()

        self.record_attn = record_attn
        self.pos_encoder = PositionalEncoding(hidden, dropout=0.1)
        encoder_layers = TransformerLayer(hidden,
                                          nhead=8,
                                          dropout=0.1,
                                          dim_feedforward=512,
                                          batch_first=True)
        self.module = TransformerEncoder(encoder_layers,
                                         layers,
                                         record_attn=record_attn)

    def forward(self, x):
        x = self.pos_encoder(x)
        output = self.module(x)
        return output

    def inference(self, x):
        return self.module(x)


class ResBlock(nn.Module):
    def __init__(self, in_channels, out_channels, stride=1):
        super(ResBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(out_channels)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_channels != out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(out_channels)
            )

    def forward(self, x):
        out = self.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = self.relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, num_classes=2):
        super(ResNet, self).__init__()
        self.in_channels = 64

        self.conv1 = nn.Conv2d(1, 64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2)
        self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512, num_classes)

    def _make_layer(self, block, out_channels, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_channels, out_channels, stride))
            self.in_channels = out_channels
        return nn.Sequential(*layers)

    def forward(self, x):
        out = self.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = self.avg_pool(out)
        out = out.view(out.size(0), -1)
        out = self.fc(out)
        return out

class InsulationScoreModule(nn.Module):
    def __init__(self, window_size=10):
        super().__init__()
        self.window_size = window_size
        self.avg_pool = nn.AvgPool2d((self.window_size, self.window_size), stride=(1, 1),padding=(0, 0))
        # self.ad_pool = nn.AdaptiveAvgPool1d(209)
        

    def forward(self, x):
        x = self.avg_pool(x)
        # x = x.squeeze(dim=-1)
        x = torch.diagonal(x, offset=self.window_size + 1, dim1=-2, dim2=-1)
        # x = self.ad_pool(x)
        return x
# class InsulationScoreModule(nn.Module):
#     def __init__(self, window_size=10, ignor_diag=2, hidden_dim=256):
#         super(InsulationScoreModule, self).__init__()
#         self.conv1 = nn.Conv2d(1, hidden_dim, kernel_size=3, stride=1, padding=1)
#         self.window_size = window_size
#         self.ignor_diag = ignor_diag - 1
#         self.relu = nn.ReLU(inplace=True)
#         self.conv2 = nn.Conv2d(hidden_dim, 1, kernel_size=(209, 1), stride=1)
#         # self.avg_pool = nn.AdaptiveAvgPool1d(209)

#     def forward(self, x):
#         mask = torch.ones_like(x)
#         mask = torch.tril(mask, self.window_size) * torch.triu(mask, -self.window_size)  # window outlier mask
#         mask_reverse = torch.tril(mask, self.ignor_diag) * torch.triu(mask, -self.ignor_diag)
#         mask_reverse = torch.where(mask_reverse == 0, 1, 0)  # ignore diag mask
#         x = x * mask * mask_reverse
#         # print(x)
#         x = x.unsqueeze(1)  # (batch_size, 1, 256, 256)
#         x = self.conv1(x)
#         x = self.relu(x)  # (batch_size, hidden_dim, 256, 256)
#         x = self.conv2(x)  # (batch_size, 1, 1, 256)
#         x = x.squeeze(1).squeeze(1)
#         # x = self.avg_pool(x)
#         return x